from flask import request, make_response, jsonify, json
from datetime import datetime, timedelta
from functools import wraps
from app import db, socketio
from app.patients.models import ImgLogo
from app.users.models import User
from app.devices.models import Device
from app.Site.models import Site
from app.patients.models import Patient
from app.rooms.models import Room
from app.zone.models import Zone
from enum import Enum
from app.logs.utils import add_log_history, construct_data, append_data_in_alert_history
from sqlalchemy import desc
from PIL import Image
from sqlalchemy import and_, or_
from app.logs.models import Logs, AdvFrameType, LogsTimeline, AlertLogsHistory
import jwt
import base64
import os
import io

class Status(Enum):
    DELETED = 0
    DRAFT = 1
    PUBLISHED = 2

class RoleEnum(Enum):
    ADMIN = (1,)
    CREATOR = (2,)
    FAN = 3

def generate_user_jwt_token(user_id, days=365):
    payload = {
        'user_id': user_id,
        'exp': datetime.utcnow() + timedelta(days=days)  
    }
    return jwt.encode(payload, os.getenv('JWT_SECRET'), algorithm='HS256')

def token_required(*roles):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            token = request.headers.get('Authorization')
            if not token:
                return make_response(jsonify({"message": "Login required!"}), 401)
            try:
                current_user = User.query.filter_by(id=jwt.decode(token.split(' ')[1], os.getenv('JWT_SECRET'),
                                                                  algorithms=['HS256'])['user_id']).first()
                if not current_user:
                    raise Exception("User not found")
                if roles and current_user.role_id not in roles:
                    return make_response(jsonify({"message": "Unauthorized access!"}), 403)
            except jwt.ExpiredSignatureError:
                return make_response(jsonify({"message": "Expired token!"}), 401)
            except jwt.InvalidTokenError:
                return make_response(jsonify({"message": "Invalid token!"}), 401)
            except Exception as e:
                return make_response(jsonify({"message": "Error decoding token: {}".format(str(e))}), 401)
            return f(current_user, *args, **kwargs)
        return wrapper
    return decorator


def create_thumbnail(image_file):
    try:
        img = Image.open(image_file)
        if img.mode in ("RGBA", "P"):
            img = img.convert("RGB")
        img.thumbnail((150, 150), Image.Resampling.LANCZOS)
        image_io = io.BytesIO()
        img_format = img.format if img.format else "PNG"
        img.save(image_io, format=img_format, optimize=True, quality=90)
        image_io.seek(0)
        return image_io, f"image/{img_format.lower()}"
    except Exception as e:
        print(f"Error creating thumbnail: {e}")
        return None, None

@socketio.on('connect')
def handle_connect():
    print('Client connected')

@socketio.on('disconnect')
def handle_disconnect():
    print('Client disconnected')

@socketio.on('alert_notification')
def send_message_to_socket(logo):
    alerts = Logs.query.filter_by(is_alert=0).all()
    if alerts:
        for alert in alerts:
            notify_text = alert.notify_text
            if notify_text:
                arr_data = json.loads(notify_text)
                arr_data['alerts_id'] = alert.alert_id
                if logo:
                    arr_data['Logo'] = logo
                socketio.emit('alert_notification', arr_data)
                print("DONEEEEEEEEEEEEEEEEEEEEEEEEEEEE")
                alert.is_alert = 1
                db.session.commit()
    else:
        print("No alerts found.")

@socketio.on('alert_accepted')
def send_message(current_user, alert_id):
    user_name = json.loads(current_user.full_name)
    first_name = user_name['first_name']
    last_name = user_name['last_name']
    user_data = {
        "user": current_user.id,
        "first_name": first_name,
        "last_name": last_name,
        "status": "Accepted",
        "alert_id": alert_id
    }
    alert_accepted = json.dumps(user_data)
    socketio.emit("alert_accepted", alert_accepted)

@socketio.on('alert_closed')
def closed_alert(current_user, alert_id):
    user_name = json.loads(current_user.full_name)
    first_name = user_name['first_name']
    last_name = user_name['last_name']
    user_data = {
        "user": current_user.id,
        "first_name": first_name,
        "last_name": last_name,
        "status": "Closed",
        "alert_id": alert_id
    }
    alert_closed = json.dumps(user_data)
    socketio.emit("alert_closed", alert_closed)

def create_alert_for_notify_zero():
    try:
        adv_frame_types = AdvFrameType.query.filter_by(notify=0).all()
        alerts_id = 0
        for adv_frame_type in adv_frame_types:
            call_action = adv_frame_type.call_action
            dt_mac = adv_frame_type.dt_mac
            if call_action == 'Single':
                call_action = "Call"
            elif call_action == 'Double':
                call_action = "Emergency"
            device = Device.query.filter_by(device_mac_address=dt_mac).first()
            site= Site.query.filter_by(site_id=device.site_id).first()
            patient = Patient.query.filter_by(device_id=device.device_id).first() if device else None
            logo = None
            staff = None
            whos_alert =""
            name = ""
            if not patient:
                staff = User.query.filter_by(device_id=device.device_id).first() if device else None
                user_name = json.loads(staff.full_name)
                first_name = user_name['first_name']
                last_name = user_name['last_name']
                name = f"{first_name} {last_name}"
                logo =get_logo_from_logotable(staff.logo_id) if staff.logo_id else (None, None)
                whos_alert = "[1]"
                flag = 1
                logo_id = staff.logo_id
                id = staff.id
            else:
                logo_id = patient.logo_id
                logo = get_logo_from_logotable(patient.logo_id) if patient.logo_id else (None, None)
                whos_alert = "[0]"
                id = patient.patient_id
                flag = 0
                name = patient.patient_name
            room = Room.query.filter_by(id=device.room_id).first() if device else None
            zone = Zone.query.filter_by(zone_id=room.zone_id).first() if device else None
            device_name = device.device_name if device else 'Unknown Device'
            room_name = room.room_name if room else 'Unknown Room'
            zone_name = zone.zone_name if zone else ''
            if adv_frame_type.flag == 0:
                alert_json = json.dumps({
                    "alert_type": call_action,
                    "Device_id": device.device_id if device else None,
                    "Device": device_name,
                    "Mac": dt_mac,
                    "Patient_name": name,
                    "Room": room_name,
                    "action": "com.example.smart_sense.PUSH_NOTIFICATION",
                    "title": "Notification Title",
                    "message": "This is the notification message",
                    "other_data": "Any other data",
                    "Zone": zone_name,
                    "Zone_id": [device.zone_id if device else None],
                    "site_id": device.site_id if device else None,
                    "whos_alert": whos_alert
                    })
                new_alert = Logs(notify_text=alert_json, is_alert=0)
                db.session.add(new_alert)
                adv_frame_type.notify = 1
                db.session.commit()
                alerts_id = new_alert.alert_id
                alert_log_history = AlertLogsHistory.query \
                    .filter(AlertLogsHistory.patient_id == id) \
                    .order_by(desc(AlertLogsHistory.id)) \
                    .first()
                with db.session.no_autoflush:
                    if alert_log_history and alert_log_history.flag != 2:
                        alert_log_history.alert_type = call_action
                        alert_log_history.flag = None
                        alert_log_history.alert_id = alerts_id
                        alert_log_array = json.loads(alert_log_history.alert_logg_array)
                        alert_log_array['alert_type'] = call_action
                        alert_log_history.alert_logg_array = json.dumps(alert_log_array)
                        repeat_alarm = LogsTimeline(alert_id=alert_log_history.id, status="Escalated", alert_type=call_action)
                        db.session.add(repeat_alarm)
                        db.session.commit()
                        append_data_in_alert_history("", alert_log_history.id, "Escalated", " "
                                                     , call_action, repeat_alarm, 0)
                    else:
                        if patient or staff:
                            alert_log_history = AlertLogsHistory(
                                alert_id=alerts_id,
                                patient_id=id,
                                alert_logg_array=construct_data(call_action, device, patient, name,
                                                                room_name, site, zone_name, room, id, logo_id),
                                zone_id=room.zone_id,
                                patient_or_staff_id=flag,
                                site_id=device.site_id
                            )
                            db.session.add(alert_log_history)
                            db.session.commit()
                            first_timeline_entry = LogsTimeline(alert_id=alert_log_history.id, status="Active",
                                                                alert_type=call_action)
                            db.session.add(first_timeline_entry)
                            db.session.commit()
                            add_log_history(call_action, device, room, site, zone_name, patient, name, room_name,
                                            alert_log_history, first_timeline_entry, flag, id, logo_id)
            send_message_to_socket(logo[0])
    except Exception as e:
        db.session.rollback()
        print(f"Error: {e}")
    finally:
        db.session.close()

def add_logo(logo):
    if logo:
        image_file, logo_content_type = create_thumbnail(logo)
        image_data = base64.b64encode(image_file.read()).decode('utf-8')
        new_logo = ImgLogo(**{"logo": image_data, "logo_content_type": logo_content_type})
        db.session.add(new_logo)
        db.session.commit()
        return new_logo.id
    return 0

def get_logo_from_logotable(id):
    img_logo = ImgLogo.query.filter_by(id=id).first()
    return base4_image(img_logo.logo), img_logo.logo_content_type

def base4_image(image):
    return base64.b64encode(base64.b64decode(image)).decode()

def delete_record(table_name, user_details, record_id, message, table_id_field):
    try:
        record = table_name.query.filter(and_(*[getattr(table_name, table_id_field) == record_id])).first()
        if not record:
            return make_response(
                jsonify({'message': 'Record not found.', "status": "error"}), 404)
        table_name.query.session.delete(record)
        table_name.query.session.commit()
        return make_response(
            jsonify({'message': f'{message} deleted successfully.', "status": "Success"}), 200)
    except Exception as e:
        return make_response(jsonify({'message': f'Error: {str(e)}', "status": "error"}), 500)

def paginate_query(query, page, per_page):
    paginated = query.paginate(page=page, per_page=per_page, error_out=False)
    return paginated.items, {
        'total_count': paginated.total,
        'total_pages': paginated.pages,
        'start_index': (page - 1) * per_page + 1,
        'end_index': min(page * per_page, paginated.total),
        'page': page,
        'per_page': per_page
    }

def paginated_query(query, search_fields, site_condition, condtion=False):
    page = request.args.get('page', 1, type=int)
    per_page = request.args.get('per_page', 10, type=int)
    site_id = request.args.get('site_id') if request.args.get('site_id') else request.args.get('customer_id')
    search = request.args.get('search')
    if not condtion:
        if not site_id or site_id == "undefined":
            return [], {'message': "site_id not found", 'status': "error"}
        elif site_condition:
            query = query.filter(site_condition == site_id)
        if search:
            search_filters = or_(
                *[field.ilike(f'%{search}%') for field in search_fields])
            query = query.filter(search_filters)
    paginated = query.paginate(page=page, per_page=per_page, error_out=False)
    return paginated.items, {
        'total_count': paginated.total,
        'total_pages': paginated.pages,
        'start_index': (page - 1) * per_page + 1,
        'end_index': min(page * per_page, paginated.total),
        'page': page,
        'per_page': per_page
    }
def single_record_query(table, record_id, id, current_user):
    if current_user.role_id == 2:
        table =table.query.filter(record_id == id).first()
        if Site.query.filter_by(site_id=table.site_id).first().customer_id == current_user.customer_id:
            return table
    return table.query.filter(and_(record_id == id, or_(table.site_id == current_user.site_id,
                                                                current_user.role_id == 1))).first()